import json
import uuid
from typing import Any, Dict, Optional, Union

from langchain_core.documents import Document

from chroma_dp import EmbeddableTextResource, Metadata


def normalize_metadata(metadata: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    if metadata is None:
        return {}
    normalized_metadata = {}
    for k, v in metadata.items():
        if isinstance(v, (str, float, bool, int)):
            normalized_metadata[k] = v
        elif isinstance(v, list):
            normalized_metadata[k] = ",".join(v)
        elif isinstance(v, dict):
            normalized_metadata[k] = json.dumps(v)
        else:
            normalized_metadata[k] = str(v)
    return normalized_metadata


def convert_chroma_emb_resource_to_lc_doc(doc: EmbeddableTextResource) -> Document:
    return Document(page_content=doc.text_chunk, metadata=doc.metadata)


def convert_lc_doc_to_chroma_resource(
    doc: Document,
    extra_metadata: Optional[Union[Dict[str, Any], Dict[str, Metadata]]] = None,
) -> EmbeddableTextResource:
    return EmbeddableTextResource(
        text_chunk=doc.page_content,
        embedding=None,
        # extra overrides doc.metadata
        metadata={
            **normalize_metadata(doc.metadata),
            **normalize_metadata(extra_metadata),
        },
        id=str(uuid.uuid4()),
    )



import json
import sys
from typing import Annotated, Optional, List, Dict, Any

import typer

from chroma_dp import EmbeddableTextResource
from chroma_dp.utils.embedding import (
    SupportedEmbeddingFunctions,
    get_embedding_function_for_name,
)
from chroma_dp.utils.chroma import remap_features


def filter_embed(
    inf: typer.FileText = typer.Argument(sys.stdin),
    batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
    embedding_function: Optional[SupportedEmbeddingFunctions] = typer.Option(
        ..., "--ef", help="The embedding function."
    ),
    embedding_model: Optional[str] = typer.Option(
        None,
        "--model",
        help="The embedding model to be used by the embedding function.",
    ),
    embed_feature: Annotated[
        str, typer.Option(help="The embedding feature.")
    ] = "embedding",
    meta_features: Annotated[
        Optional[List[str]], typer.Option(help="The metadata features.")
    ] = None,
    id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
    doc_feature: Annotated[
        str, typer.Option(help="The document feature.")
    ] = "text_chunk",
) -> None:
    _batch: Dict[str, Any] = {
        "documents": [],
        "embeddings": [],
        "metadatas": [],
        "ids": [],
    }
    _embedding_function = get_embedding_function_for_name(
        embedding_function, model=embedding_model
    )
    for line in inf:
        doc = remap_features(
            json.loads(line),
            doc_feature,
            embed_feature=embed_feature,
            meta_features=meta_features,
            id_feature=id_feature,
        )
        _batch["documents"].append(doc.text_chunk)
        _batch["metadatas"].append(doc.metadata)
        _batch["ids"].append(doc.id)
        if len(_batch["documents"]) >= batch_size:
            _batch["embeddings"] = _embedding_function(_batch["documents"])
            for d, m, i, e in zip(
                _batch["documents"],
                _batch["metadatas"],
                _batch["ids"],
                _batch["embeddings"],
            ):
                typer.echo(
                    EmbeddableTextResource(
                        text_chunk=d, metadata=m, id=i, embedding=e
                    ).model_dump_json()
                )
            _batch: Dict[str, Any] = {
                "documents": [],
                "embeddings": [],
                "metadatas": [],
                "ids": [],
            }
    if len(_batch["documents"]) > 0:
        _batch["embeddings"] = _embedding_function(_batch["documents"])
        for d, m, i, e in zip(
            _batch["documents"],
            _batch["metadatas"],
            _batch["ids"],
            _batch["embeddings"],
        ):
            typer.echo(
                EmbeddableTextResource(
                    text_chunk=d, metadata=m, id=i, embedding=e
                ).model_dump_json()
            )



import json
import sys
import uuid
from typing import Annotated, Optional, List, Dict, Any

import typer
from chromadb import EmbeddingFunction
from chromadb.api.models import Collection

from chroma_dp.utils import smart_open
from chroma_dp.utils.embedding import (
    SupportedEmbeddingFunctions,
    get_embedding_function_for_name,
)
from chroma_dp.utils.chroma import (
    CDPUri,
    get_client_for_uri,
    remap_features,
    DistanceFunction,
)


def add_to_col(
    col: Collection,
    batch: Dict[str, Any],
    upsert: bool = False,
    ef: EmbeddingFunction = None,
) -> None:
    try:
        if ef:
            batch["embeddings"] = ef(batch["documents"])
        if upsert:
            col.upsert(**batch)
        else:
            col.add(**batch)
    except Exception as e:
        raise e


def chroma_import(
    uri: Annotated[str, typer.Argument(help="The Chroma endpoint.")],
    collection: Annotated[
        Optional[str], typer.Option(help="The Chroma collection.")
    ] = None,
    inf: typer.FileText = typer.Argument(
        sys.stdin, help="Stdin input. Requires to pass `-` as the argument."
    ),
    import_file: Optional[str] = typer.Option(
        None, "--in", help="The file to use for the import instead of stdin."
    ),
    create: Annotated[
        bool, typer.Option(help="Create the Chroma collection if it does not exist.")
    ] = False,
    limit: Annotated[int, typer.Option(help="The limit.")] = -1,
    offset: Annotated[int, typer.Option(help="The offset.")] = 0,
    batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
    embedding_function: Optional[SupportedEmbeddingFunctions] = typer.Option(
        None, "--ef", help="The embedding function."
    ),
    upsert: Annotated[bool, typer.Option(help="Upsert documents.")] = False,
    embed_feature: Annotated[
        str, typer.Option(help="The embedding feature.")
    ] = "embedding",
    meta_features: Annotated[
        Optional[List[str]], typer.Option(help="The metadata features.")
    ] = None,
    id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
    doc_feature: Annotated[
        str, typer.Option(help="The document feature.")
    ] = "text_chunk",
    distance_function: Optional[DistanceFunction] = typer.Option(
        None,
        "--df",
        help="The distance function to use when creating a collection",
    ),
) -> None:
    _embedding_function = None
    if embedding_function is not None:
        _embedding_function = get_embedding_function_for_name(embedding_function)
    if uri is None:
        raise ValueError("Please provide a ChromaDP URI.")
    parsed_uri = CDPUri.from_uri(uri)
    client = get_client_for_uri(parsed_uri)
    _collection = parsed_uri.collection or collection
    _batch_size = parsed_uri.batch_size or batch_size
    _offset = parsed_uri.offset or offset
    _limit = parsed_uri.limit or limit
    _upsert = parsed_uri.upsert or upsert
    _create = parsed_uri.create_collection or create
    _batch: Dict[str, Any] = {
        "documents": [],
        "embeddings": [],
        "metadatas": [],
        "ids": [],
    }
    _distance_function = (
        distance_function or parsed_uri.distance_function or DistanceFunction.l2
    )
    if _create:
        chroma_collection = client.get_or_create_collection(
            _collection, metadata={"hnsw:space": _distance_function.value}
        )
    else:
        chroma_collection = client.get_collection(_collection)
    lc_count = 0
    with smart_open(import_file, inf) as file_or_stdin:
        for line in file_or_stdin:
            if lc_count < _offset:
                continue
            if _limit != -1 and lc_count >= _limit:
                break
            doc = remap_features(
                json.loads(line), doc_feature, embed_feature, meta_features, id_feature
            )
            _batch["documents"].append(doc.text_chunk)
            _batch["embeddings"].append(
                doc.embedding if _embedding_function is None else None
            )  # call EF?
            _batch["metadatas"].append(doc.metadata)
            _batch["ids"].append(doc.id if doc.id else uuid.uuid4())
            if len(_batch["documents"]) >= _batch_size:
                add_to_col(chroma_collection, _batch, _upsert, _embedding_function)
                _batch = {
                    "documents": [],
                    "embeddings": [],
                    "metadatas": [],
                    "ids": [],
                }
            lc_count += 1
    if len(_batch["documents"]) > 0:
        add_to_col(chroma_collection, _batch, _upsert, _embedding_function)



import typer
from chroma_dp.chroma.chroma_export import chroma_export
from chroma_dp.chroma.chroma_import import chroma_import
from chroma_dp.processor.chunk import chunk_process
from chroma_dp.processor.embed import filter_embed
from chroma_dp.huggingface import hf_import, hf_export
from chroma_dp.processor.misc.emoji_clean import emoji_clean
from chroma_dp.producer.file.pdf import pdf_import
from chroma_dp.producer.file.text import txt_import
from chroma_dp.producer.url.url_loader import url_import

app = typer.Typer(no_args_is_help=True, help="ChromaDB Data Pipes commands.")

# Import commands
import_commands = typer.Typer(no_args_is_help=True, help="Import commands.")

import_commands.command(
    name="pdf", help="Import PDF files from target dir.", no_args_is_help=True
)(pdf_import)

import_commands.command(
    name="url", help="Imports from remote url.", no_args_is_help=True
)(url_import)

import_commands.command(
    name="txt", help="Import text files from target dir.", no_args_is_help=True
)(txt_import)

# Filter commands
transform_commands = typer.Typer(no_args_is_help=True, help="Transformer commands.")
transform_commands.command(
    name="emoji-clean",
    help="Cleans emojis from documents.",
    no_args_is_help=True,
)(emoji_clean)

app.add_typer(
    import_commands, name="imp", no_args_is_help=True, help="Import Commands."
)

app.add_typer(
    transform_commands, name="tx", no_args_is_help=True, help="Filter commands."
)

# Chunk commands

app.command(
    name="chunk",
    help="Chunk embeddable resources into smaller pieces.",
    no_args_is_help=True,
)(chunk_process)

# Embed commands

app.command(
    name="embed",
    help="Generate embeddings for embeddable resources.",
    no_args_is_help=True,
)(filter_embed)

# Chroma commands
app.command(
    name="export",
    help="Export data from ChromaDB.",
    no_args_is_help=True,
)(chroma_export)

app.command(
    name="import",
    help="Import data into ChromaDB.",
    no_args_is_help=True,
)(chroma_import)


## Dataset commands


app.command(
    name="ds-get",
    help="Gets a dataset from HF.",
    no_args_is_help=True,
)(hf_import)

app.command(
    name="ds-put",
    help="Upload a dataset to HF.",
    no_args_is_help=True,
)(hf_export)

if __name__ == "__main__":
    app()



import sys
from contextlib import contextmanager
from typing import Generator, Optional, TextIO, Union, IO, Any


@contextmanager
def smart_open(
    filename: Optional[str] = None,
    stdin: TextIO = sys.stdin,
    mode: str = "r",
) -> Generator[Union[IO[Any], TextIO], None, None]:
    fh: Union[IO[Any], TextIO] = stdin
    if filename:
        fh = open(filename, mode)
    try:
        yield fh
    finally:
        if filename and isinstance(fh, IO):
            fh.close()



from typing import (
    Optional,
    Sequence,
    Any,
    Dict,
    Union,
    Protocol,
    Generator,
    TypeVar,
    Iterable,
    Generic,
)

from chromadb.api.types import Embedding
from pydantic import BaseModel, Field

C = TypeVar("C")


class ResourceFeature(BaseModel, Generic[C]):
    feature_name: str
    feature_type: C


Metadata = Dict[str, Union[str, int, float, bool]]


class EmbeddableResource(BaseModel):
    id: Optional[str] = Field(None, description="Document ID")
    metadata: Optional[Metadata] = Field(None, description="Document metadata")
    embedding: Optional[Embedding] = Field(None, description="Document embedding")

    @staticmethod
    def resource_features() -> Sequence[ResourceFeature]:
        return [
            ResourceFeature[Embedding](
                feature_name="embedding", feature_type=Embedding
            ),
            ResourceFeature[Metadata](feature_name="metadata", eature_type=Metadata),
            ResourceFeature[str](feature_name="id", feature_type=str),
        ]


class EmbeddableTextResource(EmbeddableResource):
    text_chunk: Optional[str] = Field(None, description="Document text chunk")

    @staticmethod
    def resource_features() -> Sequence[ResourceFeature]:
        return [
            ResourceFeature[str](feature_name="text_chunk", feature_type=str),
            *super().resource_features(),
        ]


D = TypeVar("D", bound=EmbeddableResource, contravariant=True)


class ChromaDocumentSourceGenerator(Protocol[D]):
    def __iter__(self) -> Generator[D, None, None]:
        ...


class CdpProducer(Protocol[D]):
    def produce(
        self, limit: int = -1, offset: int = 0, **kwargs: Dict[str, Any]
    ) -> Iterable[D]:
        ...


class CdpProcessor(Protocol[D]):
    def process(self, *, documents: Iterable[D], **kwargs: Any) -> Iterable[D]:
        ...


class CdpConsumer(Protocol[D]):
    def consume(self, *, documents: Iterable[D], **kwargs: Dict[str, Any]) -> None:
        ...



from typing import Union, Any

import datasets


def _infer_hf_type(value: Any) -> Union[datasets.Value, datasets.features.Sequence]:
    """
    Infers the Hugging Face data type from a Python type.

    Args:
        value: The value to infer the data type from.

    Returns:
        The inferred Hugging Face data type.
    """
    if isinstance(value, bool):
        return datasets.Value("bool")
    elif isinstance(value, int):
        return datasets.Value("int32")
    elif isinstance(value, float):
        return datasets.Value("float32")
    elif isinstance(value, str):
        return datasets.Value("string")
    elif isinstance(value, list):
        if all(isinstance(elem, int) for elem in value):
            return datasets.features.Sequence(feature=datasets.Value(dtype="int32"))
        elif all(isinstance(elem, float) for elem in value):
            return datasets.features.Sequence(feature=datasets.Value(dtype="float32"))
        elif all(isinstance(elem, str) for elem in value):
            return datasets.features.Sequence(feature=datasets.Value(dtype="string"))
        else:
            raise ValueError("Unsupported list type")
    else:
        raise ValueError("Unsupported type")


def int_or_none(value: Any) -> Union[int, None]:
    """
    Converts a value to an int or None.

    Args:
        value: The value to convert.

    Returns:
        The converted value.
    """
    if value is None:
        return None
    else:
        return int(value)


def bool_or_false(value: Any) -> bool:
    """
    Converts a value to a bool or None.

    Args:
        value: The value to convert.

    Returns:
        The converted value.
    """
    if value is None:
        return False
    else:
        return bool(value)



import importlib
import sys
from typing import Optional, Callable, Dict, Any, Iterable, Annotated

import typer
from langchain_community.document_loaders.recursive_url_loader import RecursiveUrlLoader

from chroma_dp import CdpProducer, EmbeddableTextResource
from chroma_dp.processor.langchain_utils import convert_lc_doc_to_chroma_resource


class BSExtractor:
    def __init__(self, url: str, max_depth: int = 2) -> None:
        try:
            _bs = importlib.import_module("bs4")
            self.bs = _bs.BeautifulSoup
        except ImportError:
            raise ImportError(
                "The bs4 package is required. Please install it with `pip install bs4`"
            )
        self.url = url
        self.max_depth = max_depth

    def __call__(self, html: str) -> str:
        return self.bs(html, "html.parser").text


class URLProducer(CdpProducer[EmbeddableTextResource]):
    def __init__(
        self,
        url: str,
        max_depth: int = 2,
        extractor: Optional[Callable[[str], str]] = None,
        batch_size: Optional[int] = 100,
    ) -> None:
        self.url = url
        self.max_depth = max_depth
        self.batch_size = batch_size
        self.extractor = extractor or BSExtractor(url, max_depth)

    def produce(
        self, limit: int = -1, offset: int = 0, **kwargs: Dict[str, Any]
    ) -> Iterable[EmbeddableTextResource]:
        loader = RecursiveUrlLoader(
            url=self.url, max_depth=self.max_depth, extractor=self.extractor
        )
        docs = loader.load()
        _start = 0
        for _offset in range(_start, len(docs), self.batch_size):
            for doc in docs[_offset : _offset + self.batch_size]:
                yield convert_lc_doc_to_chroma_resource(doc)


def url_import(
    url: Annotated[
        str,
        typer.Argument(
            ...,
            help="The URL.",
        ),
    ],
    max_depth: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--max-depth",
            "-d",
            help="The maximum depth to crawl.",
        ),
    ] = 1,
    batch_size: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--batch-size",
            "-b",
            help="The batch size to use when processing the PDF files.",
        ),
    ] = 100,
    limit: Annotated[int, typer.Option(help="The limit.")] = -1,
    offset: Annotated[int, typer.Option(help="The offset.")] = 0,
) -> None:
    """Export PDF files from a directory to ChromaDB."""
    producer = URLProducer(url=url, max_depth=max_depth, batch_size=batch_size)
    start = offset
    count = 0
    max = limit if limit > 0 else sys.maxsize
    for doc in producer.produce():
        if count < start:
            continue
        typer.echo(doc.model_dump_json())
        count += 1
        if count >= max:
            break



import os
from enum import Enum
from typing import Any, Optional

from chromadb import EmbeddingFunction
from chromadb.utils.embedding_functions import (
    ONNXMiniLM_L6_V2,
    OpenAIEmbeddingFunction,
    CohereEmbeddingFunction,
    HuggingFaceEmbeddingFunction,
    SentenceTransformerEmbeddingFunction,
)


class SupportedEmbeddingFunctions(str, Enum):
    default = "default"
    openai = "openai"
    cohere = "cohere"
    hf = "hf"
    st = "st"


def get_embedding_function_for_name(
    name: Optional[SupportedEmbeddingFunctions], **kwargs: Any
) -> EmbeddingFunction:
    if name == SupportedEmbeddingFunctions.default:
        return ONNXMiniLM_L6_V2()
    elif name == SupportedEmbeddingFunctions.openai:
        model = (
            kwargs.get("model")
            if kwargs.get("model")
            else os.environ.get("OPENAI_MODEL_NAME", "text-embedding-ada-002")
        )
        return OpenAIEmbeddingFunction(
            api_key=os.environ.get("OPENAI_API_KEY"), model_name=model
        )
    elif name == SupportedEmbeddingFunctions.cohere:
        model = (
            kwargs.get("model")
            if kwargs.get("model")
            else os.environ.get("COHERE_MODEL_NAME", "embed-english-v3.0")
        )
        return CohereEmbeddingFunction(
            api_key=os.environ.get("COHERE_API_KEY"), model_name=model
        )
    elif name == SupportedEmbeddingFunctions.hf:
        model = (
            kwargs.get("model")
            if kwargs.get("model")
            else os.environ.get(
                "HF_MODEL_NAME", "sentence-transformers/all-MiniLM-L6-v2"
            )
        )
        return HuggingFaceEmbeddingFunction(
            api_key=os.environ.get("HF_TOKEN"), model_name=model
        )
    elif name == SupportedEmbeddingFunctions.st:
        model = (
            kwargs.get("model")
            if kwargs.get("model")
            else os.environ.get("ST_MODEL_NAME", "all-MiniLM-L6-v2")
        )
        return SentenceTransformerEmbeddingFunction(
            model_name=model,
            device=os.environ.get("ST_DEVICE", "cpu"),
            normalize_embeddings=os.environ.get("ST_NORMALIZE", "True") == "True",
        )
    else:
        raise ValueError("Please provide a valid embedding function.")



import json
from typing import Annotated, Optional, List, Dict, Any
import typer
from chromadb import GetResult, Where
from chromadb.api.models import Collection
from chromadb.api.types import validate_where

from chroma_dp import EmbeddableTextResource
from chroma_dp.utils.chroma import CDPUri, get_client_for_uri


def _get_result_to_chroma_doc_list(result: GetResult) -> List[EmbeddableTextResource]:
    """Converts a GetResult to a list of ChromaDocuments."""
    docs = []
    for idx, _ in enumerate(result["ids"]):
        docs.append(
            EmbeddableTextResource(
                text_chunk=result["documents"][idx],
                embedding=result["embeddings"][idx],
                metadata=result["metadatas"][idx],
                id=result["ids"][idx],
            )
        )
    return docs


def remap_features(
    doc: EmbeddableTextResource,
    doc_feature: Optional[str] = "text_chunk",
    embed_feature: Optional[str] = "embedding",
    id_feature: str = "id",
    meta_features: Optional[List[str]] = None,
) -> Dict[str, Any]:
    """Remaps EmbeddableTextResource features to a dictionary."""

    _metas = (
        doc.metadata
        if meta_features is None
        else {
            k: doc.metadata[k]
            for k in meta_features
            if doc.metadata is not None and k in doc.metadata
        }
    )
    return {
        f"{doc_feature}": doc.text_chunk,
        f"{embed_feature}": doc.embedding,
        f"{id_feature}": doc.id,
        **(_metas if _metas is not None else {}),
    }


def read_large_data_in_chunks(
    collection: Collection, offset: int = 0, limit: int = 100, where: Where = None
) -> GetResult:
    """Reads large data in chunks from ChromaDB."""
    result = collection.get(
        where=where,
        limit=limit,
        offset=offset,
        include=["embeddings", "documents", "metadatas"],
    )
    return result


def chroma_export(
    uri: Annotated[str, typer.Argument(help="The Chroma endpoint.")],
    collection: Annotated[
        Optional[str], typer.Option(help="The Chroma collection.")
    ] = None,
    export_file: Optional[str] = typer.Option(
        None, "--out", help="Export .jsonl file."
    ),
    append: Annotated[bool, typer.Option(help="Append to export file.")] = False,
    limit: Annotated[int, typer.Option(help="The limit.")] = -1,
    offset: Annotated[int, typer.Option(help="The offset.")] = 0,
    batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
    embed_feature: Annotated[
        str, typer.Option(help="The embedding feature.")
    ] = "embedding",
    meta_features: Annotated[
        Optional[List[str]], typer.Option(help="The metadata features.")
    ] = None,
    id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
    doc_feature: Annotated[
        str, typer.Option(help="The document feature.")
    ] = "text_chunk",
    where: Annotated[Optional[str], typer.Option(help="The filter.")] = None,
) -> None:
    if uri is None:
        raise ValueError("Please provide a ChromaDP URI.")
    parsed_uri = CDPUri.from_uri(uri)
    client = get_client_for_uri(parsed_uri)
    _collection = parsed_uri.collection or collection
    _batch_size = parsed_uri.batch_size or batch_size
    _offset = parsed_uri.offset or offset
    _limit = parsed_uri.limit or limit
    _start = _offset if _offset > 0 else 0
    chroma_collection = client.get_collection(_collection)
    col_count = chroma_collection.count()
    total_results_to_fetch = min(col_count, _limit) if _limit > 0 else col_count
    _where = None
    if where:
        _where = validate_where(json.loads(where))
    if export_file and not append:
        with open(export_file, "w") as f:
            f.write("")
    for offset in range(_start, total_results_to_fetch, _batch_size):
        _results = _get_result_to_chroma_doc_list(
            read_large_data_in_chunks(
                chroma_collection,
                offset=offset,
                limit=min(total_results_to_fetch - offset, _batch_size),
                where=_where,
            )
        )
        _final_results = [
            remap_features(
                doc,
                doc_feature=doc_feature,
                embed_feature=embed_feature,
                id_feature=id_feature,
                meta_features=meta_features,
            )
            for doc in _results
        ]
        if export_file:
            with open(export_file, "a") as f:
                for _doc in _final_results:
                    f.write(json.dumps(_doc) + "\n")
        else:
            for _doc in _final_results:
                typer.echo(json.dumps(_doc))
